"""
The MIT License (MIT)

Copyright (c) 2016 Zagaran, Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

@author: Zags (Benjamin Zagorsky)
"""

import json

from django.db import models
from django.core import serializers
from itertools import chain

class BaseModel(models.Model):
    class Meta:
        abstract = True
    
    @classmethod
    def READABLE_ATTRS(cls, type_filter=None, type_exclude=None):
        return cls._attrs_filter(lambda f: f.serialize,
                                 type_filter=type_filter, type_exclude=type_exclude)
    
    @classmethod
    def WRITEABLE_ATTRS(cls, type_filter=None, type_exclude=None):
        return cls._attrs_filter(lambda f: f.serialize and f.editable,
                                 type_filter=type_filter, type_exclude=type_exclude)
    
    @classmethod
    def READONLY_ATTRS(cls, type_filter=None, type_exclude=None):
        return cls._attrs_filter(lambda f: f.serialize and not f.editable,
                                 type_filter=type_filter, type_exclude=type_exclude)
    
    @classmethod
    def _attrs_filter(cls, filter_function, type_filter=None, type_exclude=None):
        def combined_filter(f):
            if not filter_function:
                return False
            if type_filter is not None and not isinstance(f, type_filter):
                return False
            if type_exclude is not None and isinstance(f, type_exclude):
                return False
            return True
        return [f.name for f in chain(cls._meta.fields, cls._meta.many_to_many) if combined_filter(f)]
    
    @classmethod
    def create_from_json(cls, json_str):
        model = json.loads(json_str)
        params_dict = model["fields"]
        many_to_many_dict = cls.prep_many_to_many_save(params_dict)
        # _clean_json_payload pops the m2m fields
        params_dict = cls._clean_json_payload(params_dict)
        ret = cls.objects.create(**params_dict)
        # we have to do many to many field separately because obj does not exist yet otherwise
        ret.do_many_to_many_save(many_to_many_dict)
        return ret

    @classmethod
    def get_foreign_key_fields(cls):
        return {f.name: f.related_model for f in cls._meta.fields if isinstance(f, models.ForeignKey)}

    @classmethod
    def get_many_to_many_fields(cls):
        return {f.name: f.related_model for f in cls._meta.many_to_many}

    @classmethod
    def _clean_json_payload(cls, params_dict):
        writeables = cls.WRITEABLE_ATTRS()
        date_fixes = cls.WRITEABLE_ATTRS(type_filter=models.DateField, type_exclude=models.DateTimeField)
        foreign_key_fields = cls.get_foreign_key_fields()
        many_to_many_fields = cls.get_many_to_many_fields()
        for key in params_dict.keys():
            if key not in writeables or key in many_to_many_fields:
                # pop the not-writable and m2m fields
                params_dict.pop(key)
            if key in date_fixes:
                # Strip part the time portion of the ISO timestamp
                params_dict[key] = params_dict[key].split("T")[0]
            if key in foreign_key_fields:
                params_dict[key] = foreign_key_fields[key].objects.get(pk=params_dict[key])
        return params_dict

    @classmethod
    def prep_many_to_many_save(cls, params_dict):
        many_to_many_dict = {}
        many_to_many_fields = cls.get_many_to_many_fields()
        for key in params_dict.keys():
            if key in many_to_many_fields:
                many_to_many_dict[key] = many_to_many_fields[key].objects.filter(pk__in=params_dict[key])
        return many_to_many_dict

    def do_many_to_many_save(self, many_to_many_dict):
        for field, objs in many_to_many_dict.iteritems():
            getattr(self, field).set(objs)

    def update(self, update_dict=None, **kwargs):
        """ Helper method to update objects """
        if not update_dict:
            update_dict = kwargs
        update_fields = set()
        for k, v in update_dict.iteritems():
            setattr(self, k, v)
            update_fields.add(k)
        self.save(update_fields=update_fields)
    
    def to_json(self):
        """ Returns a json representation of the object """
        # Serialize takes a list; we want just a single object
        return serializers.serialize("json", [self])[1:-1]
    
    def update_from_json(self, update_json):
        model = json.loads(update_json)
        params_dict = model["fields"]
        many_to_many_dict = self.prep_many_to_many_save(params_dict)
        params_dict = self._clean_json_payload(params_dict)
        self.update(params_dict)
        self.do_many_to_many_save(many_to_many_dict)


class TimestampedModel(BaseModel):
    created_on = models.DateTimeField(auto_now_add=True, editable=False)
    updated_on = models.DateTimeField(auto_now=True, editable=False)
    
    class Meta:
        abstract = True
    
    def update(self, update_dict=None, **kwargs):
        """ Helper method to update objects """
        if not update_dict:
            update_dict = kwargs
        update_fields = {"updated_on"}
        for k, v in update_dict.iteritems():
            setattr(self, k, v)
            update_fields.add(k)
        self.save(update_fields=update_fields)



from django import template
from django.template.defaultfilters import stringfilter
from django.utils.html import format_html, mark_safe
from django.forms import DateField, DateTimeField, ModelChoiceField, ModelMultipleChoiceField
from django.db import models

import json

register = template.Library()

@register.filter(name="getattr")
def template_getattr(obj, attr_name):
    return getattr(obj, attr_name)

@register.filter(name="getitem")
def template_getitem(obj, item_name):
    return obj[item_name]

@register.filter(name="title_space")
@stringfilter
def title_space(string):
    return string.replace("_", " ").title()

@register.filter(name="get_django_field")
def template_get_django_field(obj, field):
    val = getattr(obj, field)
    if field in obj.get_many_to_many_fields():
        return val.all()
    return val

@register.simple_tag(name="angular_model")
def angular_model(model, angular_model_name, extra_params={}):
    # returns javascript that preprocesses obj before attaching it to window where angular controller can grab it
    # TODO: would like to use ng-init instead of this
    # TODO: prevent dates from showing up one day off in certain timezones
    # TODO: prepend namespace strings to window variables
    # TODO: don't rely on JS date parsing
    ret = "<script>\n"
    if model is None:
        ret += "window.%s = {};\n" % (angular_model_name)
    else:
        # cast foreign key and m2m fields to be strings
        json_ret = model.to_json()
        model_dict = json.loads(json_ret)
        fk_fields = model.get_foreign_key_fields()
        m2m_fields = model.get_many_to_many_fields()
        for field, val in model_dict["fields"].iteritems():
            if field in fk_fields:
               model_dict["fields"][field] = str(val)
            if field in m2m_fields:
                model_dict["fields"][field] = map(str, val)
        ret += "window.%s = %s;\n" % (angular_model_name, json.dumps(model_dict))
        # adds converter for datetime fields
        for field in model.READABLE_ATTRS(type_filter=models.DateField):
            # DateTimeFields are instances of DateFields
            ret += ("window.%s.fields.%s = new Date(window.%s.fields.%s);\n" %
                    (angular_model_name, field, angular_model_name, field))
    # date_fields = model.WRITEABLE_ATTRS(type_filter=models.DateField, type_exclude=models.DateTimeField)
    # date_fields = json.dumps(date_fields)
    # ret += "window.%s.date_fields = %s" % (angular_model_name, date_fields)
    ret += "</script>\n"
    return mark_safe(ret)


@register.simple_tag(name="angular_input_field")
def angular_input_field(form_field, angular_model_name, extra_params={}):
    # TODO: Integrate with angular form validation
    # TODO: Add time picker https://angular-ui.github.io/bootstrap
    # TODO: ng-options
    # TODO: angular side toLocaleDateString() for DateField
    # TODO: ng validation messages for more than just required
    # TODO: figure out what to do about angular's timezone support
    # TODO: test extra params
    # TODO: fix double datepicker on chrome
    # TODO: Investigate converting dates in angular controller http://stackoverflow.com/a/15346236/2800876
    
    try:
        form_field_value = form_field.value
        form_field.value = lambda: None
        attrs = {"ng-model": "%s.fields.%s" % (angular_model_name, form_field.name),
                 "class": "form-control"}
        if form_field.field.required:
            attrs["required"] = "true"
        if isinstance(form_field.field, DateTimeField):
            separator = extra_params.pop("silica_datetime_separator", "")
            widget1 = _get_datepicker(form_field, attrs, extra_params)
            attrs["type"] = "time"
            attrs.update(extra_params)
            widget2 = form_field.as_widget(attrs=attrs)
            return format_html(widget1 + separator + widget2)
        if isinstance(form_field.field, DateField):
            return format_html(_get_datepicker(form_field, attrs, extra_params))
        if isinstance(form_field.field, ModelChoiceField):
            pass
            #return format_html(_get_datepicker(form_field, attrs, extra_params))
        if isinstance(form_field.field, ModelMultipleChoiceField):
            pass
            #return format_html(_get_datepicker(form_field, attrs, extra_params))
        attrs.update(extra_params)
        return format_html(form_field.as_widget(attrs=attrs))
    finally:
        form_field.value = form_field_value

def _get_datepicker(form_field, attrs, extra_params):
    attrs = dict(attrs)
    calendar_button = extra_params.pop("silica_calendar_button", True)
    attrs["type"] = "date"
    attrs["placeholder"] = "yyyy-mm-dd"
    attrs["uib-datepicker-popup"] = ""
    attrs["datepicker-options"] = "dateOptions"
    attrs["is-open"] = "_calendar_widgets[%s]" % id(form_field)
    attrs["ng-click"] = "_calendar_widgets[%s]=true" % id(form_field)
    # attrs["class"] += " datepicker"
    attrs.update(extra_params)
    ret = form_field.as_widget(attrs=attrs)
    if calendar_button:
        ret = """<div class="input-group">""" + ret
        ret += """<span class="input-group-btn"><button type="button" class="btn btn-default" ng-click="_calendar_widgets[%s]=true">
                  <i class="glyphicon glyphicon-calendar"></i></button></span>""" % id(form_field)
        ret += "</div>"
    return ret
